/* The MazezaM Engine 
 * Copyright (C) 2002-2008 Malcolm Tyrrell
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */
package application;

import file.LevelFileFilter;
import file.LevelFileParseException;
import globals.Constants;
import globals.Globals;
import globals.SettingsManager;

import java.awt.Component;
import java.io.File;

import javax.swing.JFileChooser;
import javax.swing.JOptionPane;

import core.GameSwitcher;
import core.LevelCollectionInfo;

/**
 * A factory for opening dialogs used by the application.
 */
public class Dialogs {
	/**
	 * The dialogs (there should only be one of each)
	 */
	private static KeySettingsDialog keySettings_dialog;
	private static PreferencesDialog preferences_dialog;
	private static HelpDialog help_dialog;
	private static InstructionsDialog instructions_dialog;

	public static void open_keySettingsDialog() {
		if (keySettings_dialog == null)
			keySettings_dialog = new KeySettingsDialog();
		else
			keySettings_dialog.setVisible(true);
	}

	public static void open_preferencesDialog(GameSwitcher gp) {
		if (preferences_dialog == null)
			preferences_dialog = new PreferencesDialog(gp);
		else
			preferences_dialog.setVisible(true);
	}

	public static void open_helpDialog() {
		if (help_dialog == null)
			help_dialog = new HelpDialog();
		else
			help_dialog.setVisible(true);
	}

	public static void open_instructionsDialog(GameSwitcher gp) {
		if (instructions_dialog == null)
			instructions_dialog = new InstructionsDialog();
		else
			instructions_dialog.setVisible(true);
	}

	public static void open_aboutLevelsDialog(LevelCollectionInfo lli) {
		new AboutLevelsDialog(lli);
	}

	public static void close_keySettingsDialog() {
		if (keySettings_dialog != null)
			keySettings_dialog.cancel();
	}

	public static void close_preferencesDialog() {
		if (preferences_dialog != null)
			preferences_dialog.cancel();
	}

	public static void close_helpDialog() {
		if (help_dialog != null)
			help_dialog.cancel();
	}

	public static void close_instructionsDialog() {
		if (instructions_dialog != null)
			instructions_dialog.cancel();
	}

	/**
	 * A simple option dialog that checks if the user really wants to perform
	 * the action when a game is being played.
	 */
	static boolean confirmDialog(String query) {
		if (SettingsManager.app_queryForQuits) {
			String small_query = query.substring(0, 1).toLowerCase()
					+ query.substring(1);
			String[] messages = {
					Globals.text.getString("confirm_really") + small_query
							+ "?", Globals.text.getString("confirm_your_loss") };
			String[] options = { Globals.text.getString("yes_button"),
					Globals.text.getString("no_button") };
			/*
			 * Although there is a yes/no ConfirmDialog on offer, it defaults to
			 * yes. This seems wrong since pressing "Enter" (annoyingly) selects
			 * the default.
			 */
			return JOptionPane.showOptionDialog(Globals.graphics_frame,
					messages, query, JOptionPane.YES_NO_OPTION,
					JOptionPane.QUESTION_MESSAGE, null, options, Globals.text
							.getString("no_button")) == JOptionPane.YES_OPTION;
		} else
			// the user doesn't wan't to be asked, so we assume yes!
			return true;
	}

	static boolean confirmRestore(Component parent) {
		String[] text_confirmRestoreMessages = {
				Globals.text.getString("restore_really"),
				Globals.text.getString("restore_warning") };
		return JOptionPane.showConfirmDialog(parent,
				text_confirmRestoreMessages, Globals.text
						.getString("confirm_restore"),
				JOptionPane.YES_NO_OPTION, JOptionPane.QUESTION_MESSAGE) == JOptionPane.YES_OPTION;
	}

	// FIXME The default levels should be in the jar file so
	// this should just throw an exception
	public static void defaultFileLoadFailed() {
		String[] text_dialog_defaultFailed = {
				Globals.text.getString("file_load_failed_0"),
				Globals.text.getString("file_load_failed_1a")
						+ Globals.text.getString("menu_openFile")
						+ Globals.text.getString("file_load_failed_1b"),
				Globals.text.getString("file_load_failed_2") };

		JOptionPane.showMessageDialog(Globals.graphics_frame,
				text_dialog_defaultFailed, Globals.text
						.getString("default_file_failed"),
				JOptionPane.WARNING_MESSAGE);
	}

	public static final int fileLoad_retry = 0;
	public static final int fileLoad_useDefaults = 1;
	public static final int fileLoad_cancel = 2;
	public static final int fileLoad_okay = 3;

	public static int fileLoadFailed(LevelFileParseException e) {
		String[] messages = { e.getText1(), e.getText2() };
		String[] options = { Globals.text.getString("retry_button"),
				Globals.text.getString("play_defaults_button"),
				Globals.text.getString("cancel_button") };
		int response = JOptionPane.showOptionDialog(Globals.graphics_frame,
				messages, Globals.text.getString("file_load_failed_title"),
				JOptionPane.YES_NO_CANCEL_OPTION, JOptionPane.WARNING_MESSAGE,
				null, options, null);
		switch (response) {
		case JOptionPane.YES_OPTION:
			return fileLoad_retry;
		case JOptionPane.NO_OPTION:
			return fileLoad_useDefaults;
		default:
			return fileLoad_cancel;
		}
	}

	/**
	 * Returns the filePath the user choose, or null if the user cancelled.
	 */
	public static File chooseFile(LevelFileFilter[] filters, String path) {
		File file = null;
		int i;
		JFileChooser chooser = new JFileChooser(path);
		for (i = 0; i < filters.length; i++)
			chooser.addChoosableFileFilter(filters[i]);
		chooser.setDialogTitle(Globals.text.getString("file_open"));
		//chooser.setCurrentDirectory(path);
		int response = chooser.showOpenDialog(Globals.graphics_frame);
		if (response == JFileChooser.APPROVE_OPTION) {
			file = chooser.getSelectedFile();
		}
		return file;
	}

	public static void applicationAbout() {
		String[] messages = {
				Constants.text_applicationTitle,
				Constants.text_applicationVersion,
				Constants.text_applicationCopyright,
				Globals.text.getString("comments_to")
						+ Constants.text_applicationAuthorEmail };
		JOptionPane.showMessageDialog(Globals.graphics_frame, messages,
				Globals.text.getString("about_application")
						+ Constants.text_applicationTitle,
				JOptionPane.PLAIN_MESSAGE);
	}


}
