/* The MazezaM Engine 
 * Copyright (C) 2002-2008 Malcolm Tyrrell
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */
package core;

import globals.SettingsManager;

import java.util.LinkedList;

import tessellation.TessellationManager;

/**
 * Stores playerActions and supports undoing (via reapplying actions to a resetted level).
 */
public class GameUndoManager {
	// Needs to be < -1 (See add_playerAction)
	public static final int unbounded = -99;
	public static int[] availableUndoLevels = { 0, 1, 2, 4, 8, 16, 32, 64, 128, 256,
			unbounded };

	LinkedList<PlayerAction> action_list;
	LevelInterface the_level;
	
	/**
	 * Counts the number of successfully applied actions, so that 
	 * it can be judged whether it's safe to abandon a level.
	 */
	int move_count;

	/**
	 * Create an undo manager for the given level.
	 * 
	 * @param l
	 *            the level to manage undos for.
	 */
	GameUndoManager(LevelInterface l) {
		action_list = new LinkedList<PlayerAction>();
		// Have to make sure the Tessellation is set correctly before a clone
		TessellationManager.set_tessellation(l.get_gameInfo().get_tessellation());
		the_level = (LevelInterface) l.clone();
		move_count = 0;
	}

	/**
	 * Return true if there's an available undo
	 */
	boolean can_undo() {
		return !action_list.isEmpty();
	}

	/**
	 * Have undoable PlayerActions occurred?
	 */
	boolean can_abandonWithoutQuery() {
		return move_count == 0;
	}
	
	/**
	 * Reset the undo structure
	 */
	void reset_list(LevelInterface l) {
		action_list.clear();
		// Have to make sure the Tessellation is set correctly before a clone
		TessellationManager.set_tessellation(l.get_gameInfo().get_tessellation());
		the_level = (LevelInterface) l.clone();
		move_count = 0;
	}

	/**
	 * Add a player action to the undo manager.
	 */
	void add_playerAction(PlayerAction a) {
		if (SettingsManager.app_maxUndos != 0) {
			action_list.add(a);
			if (action_list.size() == SettingsManager.app_maxUndos + 1) {
				action_list.getFirst().attemptAction(the_level);
				action_list.removeFirst();
			}
		}
		move_count = move_count + 1;
	}

	/**
	 * Return the level which results from an undo.
	 */
	LevelInterface undo() {
		// Don't want to repeat the very last action!
		action_list.removeLast();
		// Have to make sure the Tessellation is set correctly before a clone
		TessellationManager.set_tessellation(the_level.get_gameInfo().get_tessellation());
		// Need to keep the old level so that I can undo again.
		LevelInterface l = (LevelInterface) the_level.clone();
		// apply the actions to the level
		for (PlayerAction a : action_list) {
			a.attemptAction(l);
		}
		move_count = move_count - 1;
		return l;
	}
}