/* The MazezaM Engine 
 * Copyright (C) 2002-2008 Malcolm Tyrrell
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */
package file;

import globals.Constants;
import globals.Globals;

import java.io.File;
import java.io.FileInputStream;
import java.io.FileNotFoundException;
import java.io.IOException;
import java.net.MalformedURLException;
import java.net.URL;
import java.util.LinkedList;

import application.Dialogs;

import core.AvailableGameManager;
import core.LevelCollection;

/**
 * A facade to the level loading system.
 */
public class FileManager {
	/** used to open a file dialog at sensible locations. */
	static String last_file_loaded = sokoban.Solver.LEVEL_INPUT;
	/**
	 * used to remember whether we were playing the defaults last time the
	 * application was run.
	 */
	static boolean playing_defaults;

	/**
	 * If we're in applet mode, this is a level provided by the applet context
	 */
	static String levels_at_url;

	/**
	 * Create a file manager with f as the last file used.
	 */
	public FileManager() {
	}

	LevelFileParseException the_error;

	/**
	 * Returns an array of the support levelFileParsers.
	 */
	private LevelFileParser[] get_levelFileParsers() {
		LevelParser[] lpa = AvailableGameManager.get_levelParsers();
		/*
		 * There is one extra parser for "xml" files.
		 */
		LevelFileParser[] lfpa = new LevelFileParser[lpa.length + 1];
		for (int i = 0; i < lpa.length; i++) {
			lfpa[i] = new StandardLevelFileParser(lpa[i]);
		}
		lfpa[lpa.length] = new XmlLevelFileParser(lpa);
		return lfpa;
	}

	/**
	 * Returns an array of the supported levelFileFilters.
	 */
	private LevelFileFilter[] get_levelFileFilters() {
		LevelParser[] lpa = AvailableGameManager.get_levelParsers();
		/*
		 * There are two filters more than there are level parsers. One is for
		 * "xml" files and the other is a filter which accepts all supported
		 * files.
		 */
		String[] extensions = new String[lpa.length + 1];
		LevelFileFilter[] lffa = new LevelFileFilter[lpa.length + 2];
		for (int i = 0; i < lpa.length; i++) {
			extensions[i] = lpa[i].getExtension();
			lffa[i] = lpa[i].get_filter();
		}

		lffa[lpa.length] = new LevelFileFilter("xml", "XML"
				+ Globals.text.getString("file_filter_files")
				+ Globals.text.getString("file_filter_multi"));
		extensions[lpa.length] = "xml";

		lffa[lpa.length + 1] = new LevelFileFilter(extensions, Globals.text
				.getString("file_filter_supported"));
		return lffa;
	}

	/**
	 * Find a parser which claims to be able to handle files of this filename
	 */
	private LevelFileParser suitableParser(String filename) {
		LevelFileParser[] lfpa = get_levelFileParsers();
		for (int i = 0; i < lfpa.length; i++) {
			if (lfpa[i].get_filter().accept(filename))
				return lfpa[i];
		}
		// FIXME If I can't find a match, assume the first parser.
		return lfpa[0];
	}

	/**
	 * Launches a file open dialog and attempts to load the chosen file
	 */
	public LinkedList<LevelCollection> openFile()
			throws LevelFileParseException {
		File file = Dialogs
				.chooseFile(get_levelFileFilters(), last_file_loaded);
		if (file != null)
			return loadLevelsFromFileSystem(file);
		else
			return null;
	}

	public LinkedList<LevelCollection> reloadLastFile() {
		LinkedList<LevelCollection> the_levels = null;
		try {
		     the_levels = parseLevelListFromFileSystem(new File(
				     last_file_loaded));
		} catch (LevelFileParseException e) {
		    throw new RuntimeException(e);
		}
		return the_levels;
	}

	public LinkedList<LevelCollection> loadAppletLevels() throws LevelFileParseException {
		LinkedList<LevelCollection> the_levels = null;
		if (levels_at_url == null)
			the_levels = loadDefaults();
		else
			the_levels = parseLevelListAtUrl(levels_at_url);
		return the_levels;
	}

	/**
	 * Attempt to load a level file
	 */
	private LinkedList<LevelCollection> loadLevelsFromFileSystem(File file)
			throws LevelFileParseException {
		LinkedList<LevelCollection> the_levels = null;
		// Because the dialog for failure offers a retry, we need to wrap
		// it in a loop.
		int response = Dialogs.fileLoad_retry;
		while (response == Dialogs.fileLoad_retry) {
			try {
				the_levels = parseLevelListFromFileSystem(file);
				// update the fact that this path becomes the default.
				playing_defaults = false;
				response = Dialogs.fileLoad_okay;
			} catch (LevelFileParseException e) {
				response = Dialogs.fileLoadFailed(e);
				if (response == Dialogs.fileLoad_useDefaults) {
					the_levels = loadDefaults();
					playing_defaults = true;
				}
			}
		}
		try {
			last_file_loaded = file.getCanonicalPath();
		} catch (IOException e) {
		}
		return the_levels;
	}

	/**
	 * Perform the transition to the gameStart or fileError state by loading the
	 * default file in the containing jar file
	 */
	public LinkedList<LevelCollection> loadDefaults()
			throws LevelFileParseException {
		LinkedList<LevelCollection> the_levels;
		the_levels = parseLevelListInJar(Constants.file_xml_levels_in_jar);
		playing_defaults = true;
		return the_levels;
	}

	/**
	 * Attempt to parse the levels in the file with the jar file
	 */
	private LinkedList<LevelCollection> parseLevelListInJar(String path)
			throws LevelFileParseException {
		LinkedList<LevelCollection> levels = null;
		LevelFileParseException error = null;
		try {
			// By using a class loader to get at the level file, it will work in
			// a jar file.
			URL levels_url = getClass().getClassLoader().getResource(path);

			if (levels_url != null) {
				levels = suitableParser(path).parseLevels(levels_url.getFile(),
						levels_url.openStream());
			} else {
				error = new LevelFileParseException(Globals.text
						.getString("error_fileNotFound"), path);
			}
		} catch (LevelFileParseException e) {
			e.setFileName(path);
			error = e;
		} catch (IOException e) {
			error = new LevelFileParseException(e.getMessage(), path);
		}
		if (error == null)
			return levels;
		else
			throw error;
	}

	/**
	 * Attempt to parse the levels in the file at the url
	 */
	private LinkedList<LevelCollection> parseLevelListAtUrl(String url_text)
			throws LevelFileParseException {
		LinkedList<LevelCollection> levels = null;
		LevelFileParseException error = null;
		try {
			URL levels_url = new URL(url_text);

			if (levels_url != null) {
				// determine a nice display name
				String name = levels_url.getPath();
				name = name.substring(name.lastIndexOf("/")+1);
				levels = suitableParser(url_text).parseLevels(
						name, levels_url.openStream());
			} else {
				// Don't know the conditions under which this happens... but
				// it does happen.
				throw new Error(
						"Wasn't able to construct a valid url object from "
								+ url_text);
			}
		} catch (LevelFileParseException e) {
			e.setFileName(url_text);
			error = e;
		} catch (MalformedURLException e) {
			error = new LevelFileParseException(Globals.text
					.getString("error_badUrl"), url_text);
		} catch (IOException e) {
			error = new LevelFileParseException(e.getMessage(), url_text);
		} catch (SecurityException e) {
			error = new LevelFileParseException(e.getMessage(), url_text);
		}
		if (error == null)
			return levels;
		else
			throw error;
	}

	/**
	 * Attempt to parse the levels in the file in the file system
	 */
	private LinkedList<LevelCollection> parseLevelListFromFileSystem(File file)
			throws LevelFileParseException {
		LinkedList<LevelCollection> levels = null;
		LevelFileParseException error = null;
		try {
			levels = suitableParser(file.toString()).parseLevels(
					file.getName(), new FileInputStream(file));
		} catch (LevelFileParseException e) {
			e.setFileName(file.toString());
			error = e;
		} catch (FileNotFoundException e) {
			error = new LevelFileParseException(Globals.text
					.getString("error_fileNotFound"), file.toString());
		}
		if (error == null)
			return levels;
		else
			throw error;
	}

	public static void accessSettings() {
//		last_file_loaded = Globals.settings.getProperty("app_lastFileLoaded");
		playing_defaults = Boolean.parseBoolean(Globals.settings.getProperty(
				"playing_defaults", "true"));
		levels_at_url = Globals.settings.getProperty("levels_at_url",null);
	}

	public static void updateSettings() {
		if (last_file_loaded != null)
			Globals.settings
					.setProperty("app_lastFileLoaded", last_file_loaded);
		Globals.settings.setProperty("playing_defaults", Boolean
				.toString(playing_defaults));
	}
}
