/* The MazezaM Engine 
 * Copyright (C) 2002-2008 Malcolm Tyrrell
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */
package file;

import globals.Globals;

import java.io.IOException;
import java.io.InputStream;
import java.util.Iterator;
import java.util.LinkedList;

import javax.xml.parsers.FactoryConfigurationError;
import javax.xml.parsers.ParserConfigurationException;
import javax.xml.parsers.SAXParser;
import javax.xml.parsers.SAXParserFactory;

import level.Level;

import org.xml.sax.Attributes;
import org.xml.sax.Locator;
import org.xml.sax.SAXException;
import org.xml.sax.SAXParseException;
import org.xml.sax.helpers.DefaultHandler;

import core.LevelCollection;

/**
 * A call-back interface used to build a LevelList based on events generated by
 * a SAXParser object.
 * 
 * The parsing work is done in the body of the parseLevelList method.
 */
class XmlLevelFileHandler extends DefaultHandler {
	private int state;

	static boolean inParserSection;

	// This object can be in the following states while parsing is underway.
	static final int state_initial = 0;
	static final int state_inLevelCollection = 1;
	static final int state_inLevel = 2;
	static final int state_inLevelLine = 3;
	static final int state_inTitle = 4;
	static final int state_inDescription = 5;
	static final int state_inEMail = 6;
	static final int state_inUrl = 7;

	// The element names of a MultiGameLevels file
	static final String element_levelCollection = "LevelCollection";
	static final String element_level = "Level";
	static final String element_levelLine = "L";
	static final String element_title = "Title";
	static final String element_description = "Description";
	static final String element_eMail = "Email";
	static final String element_url = "Url";

	// The following will be combined with a game title to give an element
	// name like "SokobanLevels";
	//static String element_parserSection;
	static final String parserSectionSuffix = "Levels";

	static final String att_title = "Id";
	static final String att_copyright = "Copyright";

	private LinkedList<LevelCollection> the_levels_list;
	private LinkedList<Level> the_levels;
	private LinkedList<String> the_lines;
	private int level_count;
	private Locator the_locator;
	private LevelParser[] the_parsers;
	private int parser_index;

	private String collectionTitle;
	private String collectionCopyright;
	private String description;
	private String email;
	private String url;

	private String level_title;
	private String level_author;
	private int levelDepth;

	// Because XML parsing can break character data into substrings, we need to
	// store them before joining them back together.
	private LinkedList<String> temp_strings;

	/**
	 * Requires a LevelParser to parse the contents of level lines
	 */
	XmlLevelFileHandler(LevelParser[] lpa) {
		the_parsers = lpa;
		// element_parserSection = lp.getGameTitle() + parserSectionSuffix;
	}

	/**
	 * Parse an input stream as a MultiGameLevels xml file.
	 */
	LinkedList<LevelCollection> parseLevelList(InputStream is)
			throws LevelFileParseException {
		the_levels_list = new LinkedList<LevelCollection>();
		LevelFileParseException the_error = null;
		// the_locator = new LocatorImpl();
		setDocumentLocator(the_locator);
		collectionTitle = null;
		collectionCopyright = null;
		description = null;
		email = null;
		url = null;
		inParserSection = false;
		try {
			// This is where the parsing work is actually done.
			SAXParser saxparser = SAXParserFactory.newInstance().newSAXParser();
			saxparser.parse(is, this);
		} catch (SAXParseException e) {
			// If the SAXParseException was thrown by our level parsing in
			// the EndElement method, it will encapsulate our error
			// Otherwise, use the details of the SAX error to build our error.
			Exception ex = e.getException();
			if ((ex != null) && (ex instanceof LevelParseException))
				the_error = new XmlLevelFileParseException(
						(LevelParseException) ex, e.getLineNumber());
			else
				the_error = new LevelFileParseException(e.getMessage(), e
						.getLineNumber());
		} catch (FactoryConfigurationError e) {
			throw new Error("XML system error: " + e);
		} catch (ParserConfigurationException e) {
			throw new Error("XML system error: " + e);
		} catch (SAXException e) {
			the_error = new LevelFileParseException(e.getMessage());
		} catch (IOException e) {
			the_error = new LevelFileParseException(e.getMessage());
		}
		if ((the_error == null) && (the_levels.size() == 0))
			the_error = new LevelFileParseException(
			// FIXME
					Globals.text.getString("error_noLevelsFound0")
							+ Globals.text.getString("error_noLevelsFound1"));
		if (the_error == null)
			return the_levels_list;
		else
			throw the_error;
	}

	public void startDocument() {
		state = state_initial;
	}
	
	private boolean identifyParserSection(String s) {
		for (int i = 0; i < the_parsers.length; i++) {
			if (s.equals(the_parsers[i].getGameTitle() + parserSectionSuffix)) {
				parser_index = i;
				return true;
			}
		}
		return false;
	}
	
	public void startElement(String uri, String localName, String qName,
			Attributes atts) throws SAXParseException {
		String atvalue;
		if (identifyParserSection(qName)) {
			inParserSection = true;
			the_levels = new LinkedList<Level>();
		} else if (inParserSection) {
			if (qName.equals(element_levelCollection)) {
				if (state != state_initial)
					throw startError(element_levelCollection);
				// process attributes
				atvalue = atts.getValue(att_copyright);
				if (atvalue != null)
					collectionCopyright = atvalue;
				else
					collectionCopyright = "";
				state = state_inLevelCollection;
				level_count = 0;
			} else if (qName.equals(element_level)) {
				if (state != state_inLevelCollection)
					throw startError(element_level);

				// process attributes
				atvalue = atts.getValue(att_title);
				if (atvalue != null)
					level_title = atvalue;
				else
					level_title = "";
				atvalue = atts.getValue(att_copyright);
				if (atvalue != null)
					level_author = atvalue;
				else
					level_author = collectionCopyright;
				atvalue = atts.getValue("depth");
				if (atvalue != null) {
					System.out.println(atvalue);
					levelDepth = Integer.parseInt(atvalue);
				} else
					levelDepth = 0;

				state = state_inLevel;
				the_lines = new LinkedList<String>();
			} else if (qName.equals(element_levelLine)) {
				if (state != state_inLevel)
					throw startError(element_levelLine);
				state = state_inLevelLine;
				temp_strings = new LinkedList<String>();
			} else if (qName.equals(element_title)) {
				if (state != state_initial)
					throw startError(element_title);
				state = state_inTitle;
				temp_strings = new LinkedList<String>();
			} else if (qName.equals(element_description)) {
				if (state != state_initial)
					throw startError(element_description);
				state = state_inDescription;
				temp_strings = new LinkedList<String>();
			} else if (qName.equals(element_eMail)) {
				if (state != state_initial)
					throw startError(element_eMail);
				state = state_inEMail;
				temp_strings = new LinkedList<String>();
			} else if (qName.equals(element_url)) {
				if (state != state_initial)
					throw startError(element_url);
				state = state_inUrl;
				temp_strings = new LinkedList<String>();
			}
		}
	}

	public void characters(char[] ch, int start, int length) {
		if (inParserSection) {
			if (state == state_inLevelLine) {
				// because XML parsing can break character data into substrings,
				// we
				// store these substrings and flatten them into a single string
				// later.
				temp_strings.add(new String(ch, start, length));
			} else if (state == state_inTitle)
				temp_strings.add(new String(ch, start, length));
			else if (state == state_inDescription)
				temp_strings.add(new String(ch, start, length));
			else if (state == state_inEMail)
				temp_strings.add(new String(ch, start, length));
			else if (state == state_inUrl)
				temp_strings.add(new String(ch, start, length));
		}
	}

	public void endElement(String uri, String localName, String qName)
			throws SAXParseException {
		if (identifyParserSection(qName)) {
			if (!inParserSection)
				throw endError(the_parsers[parser_index].getGameTitle() + parserSectionSuffix);
			inParserSection = false;
			the_levels_list.add(new LevelCollection(the_levels,collectionTitle, collectionCopyright,description, email, url));
		}
		if (inParserSection) {
			if (qName.equals(element_levelLine)) {
				if (state != state_inLevelLine)
					throw endError(element_levelLine);
				state = state_inLevel;
				the_lines.add(flatten(temp_strings));
			} else if (qName.equals(element_level)) {
				if (state != state_inLevel)
					throw endError(element_level);
				state = state_inLevelCollection;
				try {
					the_levels.add(the_parsers[parser_index].parseLevel(
							level_title, level_author, the_lines, level_count, 
							levelDepth));
					level_count = level_count + 1;
				} catch (LevelParseException e) {

					throw new SAXParseException("", the_locator, e);
				}
			} else if (qName.equals(element_levelCollection)) {
				if (state != state_inLevelCollection)
					throw endError(element_levelCollection);
				state = state_initial;
			} else if (qName.equals(element_title)) {
				if (state != state_inTitle)
					throw endError(element_title);
				collectionTitle = flatten(temp_strings);
				state = state_initial;
			} else if (qName.equals(element_description)) {
				if (state != state_inDescription)
					throw endError(element_description);
				description = flatten(temp_strings);
				state = state_initial;
			} else if (qName.equals(element_eMail)) {
				if (state != state_inEMail)
					throw endError(element_eMail);
				email = flatten(temp_strings);
				state = state_initial;
			} else if (qName.equals(element_url)) {
				if (state != state_inUrl)
					throw endError(element_url);
				url = flatten(temp_strings);
				state = state_initial;
			}
		}
	}

	public void setDocumentLocator(Locator l) {
		the_locator = l;
	}

	private SAXParseException startError(String e) {
		return new SAXParseException(Globals.text.getString("error_xml_start")
				+ e + Globals.text.getString("error_xml_end"), the_locator);
	}

	private SAXParseException endError(String e) {
		return startError("/" + e);
	}

	private String flatten(LinkedList<String> sl) {
		Iterator<String> sli;
		StringBuffer sb = new StringBuffer();
		for (sli = sl.iterator(); sli.hasNext();) {
			sb.append(sli.next());
		}
		return sb.toString();
	}
}
