/* The MazezaM Engine 
 * Copyright (C) 2002-2008 Malcolm Tyrrell
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */
package level;

import java.util.Iterator;
import java.util.LinkedList;

import tessellation.Direction;
import tessellation.EdgeIterator;
import tessellation.NeighbourIterator;
import tessellation.PosDirBoolTuple;
import tessellation.Position;
import tessellation.TessellationManager;


/**
 * A shape consisting of one type of unit which must in a connected group but
 * may contain holes
 */
public class Chunk extends UnitShape {

	/**
	 * How far between the corner of the tile and the midpoint of the two
	 * adjacent corners should the inner corner be placed.
	 */
	protected double innercorner_proprotion = 0;

	/**
	 * Constructs a chunk from an array of strings of unit characters
	 */
	Chunk(LinkedList<Position> positions, char u, UnitPaint up) {
		Iterator<Position> pli;
		Position pos;
		int x, y;
		chunk_positions = positions;
		unit_paint = up;
		chunk_unit = u;
		for (pli = positions.iterator(); pli.hasNext();) {
			pos = pli.next();
			x = pos.array_x();
			y = pos.array_y();
			if (x >= greatest_x)
				greatest_x = x;
			else if (x < least_x)
				least_x = x;
			if (y >= greatest_y)
				greatest_y = y;
			else if (y < least_y)
				least_y = y;
		}
	}

	/**
	 * Constructs a 1-unit high chunk from a string of unit characters
	 */
	public Chunk(String s, UnitPaint up) {
		String[] string_array = new String[1];
		chunk_positions = new LinkedList<Position>();
		unit_paint = up;
		string_array[0] = s;
		least_x = 0;
		greatest_x = s.length();
		least_y = 0;
		greatest_y = 0;
		for (int x = 0; x < s.length(); x++)
			if (!Unit.charIsEmpty(s.charAt(x))) {
				// FIXME using newPosition(x,y) is dodgy when you don't know what the tessellation is!
				chunk_positions.add(TessellationManager.newPosition(x, 0));
				chunk_unit = s.charAt(x);
			}
	}

	char at(Position pos) {
		if (chunk_positions.contains(pos))
			return chunk_unit;
		else
			return Unit.empty;
	}

	LinkedList<Position> nonEmptyPositions() {
		return chunk_positions;
	}

	/**
	 * Builds a description of the outline of the contents of the chunk starting
	 * at position pos, edge ev, travelling clockwise
	 */
	/*
	 * Consider the following arrangement of units: 12 34 and say we are
	 * considering 3s north edge. We know 1 is empty and 3 is non-empty. Which
	 * edge we pick next depends on whether 2 or 4 are empty. If 4 is empty, we
	 * pick 3s east edge. If 4 is full, but 2 is empty, we pick 4s north edge.
	 * If 4 and 2 are full, but 1 is empty (which it must be), we pick 2s west
	 * edge. In general, we consider the neighbours around the target point of
	 * the current edge vector in an anticlockwise direction looking for the end
	 * of a continuous chain of filled-in blocks.
	 * 
	 * To isolate the algorithm from the details of the geometry, we leave the
	 * task of choosing which edge and position to consider next to the
	 * Direction class.
	 */
	private LinkedList<PosDirBoolTuple> findOutline(Position start_pos,
			Direction start_edge) {
		if (!(start_pos.hasNeighbour(start_edge) || empty(start_pos
				.neighbour(start_edge))))
			throw new Error("Invalid exposed position provided!");

		LinkedList<PosDirBoolTuple> pdl = new LinkedList<PosDirBoolTuple>();
		pdl.add(new PosDirBoolTuple(start_pos, start_edge, true));

		Position current_pos = start_pos;
		Direction current_edge = start_edge.nextEdgeAt(current_pos);

		Position proposed_pos = current_pos.neighbour(current_edge);
		Direction proposed_edge = current_edge.opposite().nextEdgeAt(
				proposed_pos);

		while (!(current_pos.equals(start_pos) && current_edge
				.equals(start_edge))) {
			if (empty(proposed_pos)) {
				pdl.add(new PosDirBoolTuple(current_pos, current_edge, true));
				current_edge = current_edge.nextEdgeAt(current_pos);
			} else {
				pdl.add(new PosDirBoolTuple(current_pos, current_edge, false));
				current_edge = proposed_edge;
				current_pos = proposed_pos;
			}
			proposed_pos = current_pos.neighbour(current_edge);
			proposed_edge = current_edge.opposite().nextEdgeAt(proposed_pos);
		}

		return pdl;
	}

	/**
	 * Returns a list of outlines, which are lists of PosDirBoolTuple; The first
	 * element in the list is the outer outline and the rest are for holes.
	 */
	/*
	 * The algorithm looks for positions which are exposed on an edge, but which
	 * do not lie on an outline we have yet encountered. When it finds one it
	 * adds the outline at that position and edge to the main list.
	 */
	private OutlinePair findOutlines() {
		LinkedList<PosDirBoolTuple> outer = new LinkedList<PosDirBoolTuple>();
		LinkedList<LinkedList<PosDirBoolTuple>> holes = new LinkedList<LinkedList<PosDirBoolTuple>>();
		Iterator<Position> pli;
		//LinkedList<Position> pl;
		Position p;
		PosDirBoolTuple pdp;

		EdgeIterator ei;
		Direction dir;

		for (pli = nonEmptyPositions().iterator(); pli.hasNext();) {
			p = pli.next();
			for (ei = new EdgeIterator(p); ei.hasNext();) {
				dir = ei.next();
				pdp = new PosDirBoolTuple(p, dir, true);
				if (empty(p.neighbour(dir)) && !outer.contains(pdp)
						&& notInListList(holes, pdp)) {
					if (outer.isEmpty()) {
						if (isOnOutside(p.neighbour(dir)))
							outer = findOutline(p, dir);
						else
							holes.add(findOutline(p, dir));
					} else
						holes.add(findOutline(p, dir));
				}
			}
		}
		return new OutlinePair(outer, holes);
	}

	/**
	 * Is this provided edge and position on the outside.
	 * 
	 * We do a walk at pos to see if we can leave the bounds of the shape.
	 */
	private boolean isOnOutside(Position pos) {
		if (!empty(pos))
			throw new Error("Non empty position provided");
		return isOnOutside_rec(pos, new LinkedList<Position>());
	}

	private boolean isOnOutside_rec(Position pos, LinkedList<Position> pl) {
		boolean foundOutside = false;
		NeighbourIterator ni;
		Position neighbour;
		pl.add(pos);
		if (!inside(pos))
			return true;
		for (ni = new NeighbourIterator(pos); ni.hasNext();) {
			neighbour = ni.next();
			if (empty(neighbour) && !pl.contains(neighbour))
				foundOutside = foundOutside || isOnOutside_rec(neighbour, pl);
		}
		return foundOutside;
	}

	private boolean inside(Position pos) {
		int x = pos.array_x();
		int y = pos.array_y();
		return (x >= least_x) && (x <= greatest_x) && (y >= least_y)
				&& (y <= greatest_y);
	}

	/**
	 * Returns false if the pdl is in one of the lists in pdll
	 */
	private boolean notInListList(LinkedList<LinkedList<PosDirBoolTuple>> pdll,
			PosDirBoolTuple pdl) {
		Iterator<LinkedList<PosDirBoolTuple>> pdlli;
		for (pdlli = pdll.iterator(); pdlli.hasNext();) {
			if (pdlli.next().contains(pdl))
				return false;
		}
		return true;
	}

	void paintInto(UnitGraphics g, Position rel, int superSelection) {
		OutlinePair op = findOutlines();
		LinkedList<PosDirBoolTuple> outer = op.outer_outline;
		LinkedList<LinkedList<PosDirBoolTuple>> holes = op.the_holes;
		Iterator<LinkedList<PosDirBoolTuple>> pdlli;
		Iterator<PosDirBoolTuple> pdli;
		PosDirBoolTuple pdp;
		// adjust the returned outline lists to take account of rel
		for (pdli = outer.iterator(); pdli.hasNext();) {
			pdp = pdli.next();
			pdp.pos = pdp.pos.plus(rel);
		}
		for (pdlli = holes.iterator(); pdlli.hasNext();)
			for (pdli = pdlli.next().iterator(); pdli.hasNext();) {
				pdp = pdli.next();
				pdp.pos = pdp.pos.plus(rel);
			}
		g.setPaint(unit_paint.getPaint());
		g.fill(g.edgePathWithHoles(op));
		if (isSelected(superSelection)) {
			g.setPaint((new UnitPaint_Selection(selector)).getPaint());
			g.fill(g.edgeOutlineOutline(op));
		} /* else { // Add a darkened outline
			g.setPaint(new Color(50,50,50,150));
			g.fill(g.edgeOutlineOutline(op));
		} */
	}

	public Object clone() {
		Chunk new_instance = (Chunk) super.clone();
		new_instance.chunk_unit = chunk_unit;
		new_instance.chunk_positions = (LinkedList<Position>) chunk_positions
				.clone();
		new_instance.least_x = least_x;
		new_instance.least_y = least_y;
		new_instance.greatest_x = greatest_x;
		new_instance.greatest_y = greatest_y;
		return new_instance;
	}

	private char chunk_unit;
	private LinkedList<Position> chunk_positions;
	private int least_x;
	private int least_y;
	private int greatest_x;
	private int greatest_y;
}
