/* The MazezaM Engine 
 * Copyright (C) 2002-2008 Malcolm Tyrrell
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */
package level;

import general.ObservableProxy;
import general.Pair;

import java.awt.Dimension;
import java.awt.Graphics2D;
import java.awt.Insets;
import java.awt.Point;
import java.awt.Rectangle;
import java.util.Iterator;
import java.util.LinkedList;
import java.util.Observer;

import tessellation.Direction;
import tessellation.DirectionIterator;
import tessellation.Position;
import tessellation.Tessellation;
import tessellation.TessellationManager;
import core.LevelInterface;

/**
 * A class containing a level (an individual puzzle).
 */
public abstract class Level extends UnitObject implements LevelInterface {
	protected String title;
	protected String copyright;
	protected UnitShapeGroup unitshape_group;
	protected FloorGroup floor_group;
	protected ObservableProxy ob_proxy;
	/*
	 * we may need to calculate the rectangular area of the screen when some
	 * blocks are moved, so we obtain a RectangleCalculator object from the
	 * UnitGraphics we last used. This is allowed to be null!
	 */
	// private RectangleCalculator rectangle_calculator;
	private UnitGraphics unit_graphics;

	/**
	 * Constructor for a level
	 * 
	 * @param t
	 *            the title of the level
	 * @param c
	 *            the copyright info of the level
	 * @param start_pos
	 *            a list of starting positions
	 */
	public Level(String t, String c, LinkedList<Position> start_pos) {
		title = t;
		copyright = c;
		unitshape_group = new UnitShapeGroup();
		floor_group = new FloorGroup();
		unit_paint = new UnitPaint_Floor(0);
		ob_proxy = new ObservableProxy();
		Iterator<Position> pi;
		int i = 0;
		for (pi = start_pos.iterator(); pi.hasNext();) {
			unitshape_group
					.add(new UnitShapePosPair(new Person(i++), pi.next()));
		}
		// FIXME Testing hack
		unitshape_group.selectNext(0);
	}

	public boolean attempt_direction(int p, Direction dir) {
		UnitShapePosList uspl = unitshape_group.selected(0);
		return push(uspl, dir, Resistance.defaultStrength) > 0;
	}

	public boolean attempt_advanceSelection(int s) {
		UnitShapeBounds bounds = unitshape_group.selectNext(s);
		if (bounds != null) {
			setChanged();
			if (unit_graphics != null)
				notifyObservers(unit_graphics.calculateRectangle(bounds));
			else
				notifyObservers();
			return true;
		} else
			return false;
	}

	public boolean is_completed() {
		return false;
	}

	public boolean is_failed() {
		return false;
	}

	/**
	 * Attempt to push the shapes at position pos in direction dir, using the
	 * given strength.
	 * 
	 * returns the number of shapes moved by the push
	 */
	// Note that this doesn't prevent the block leaving the level.
	protected int push(UnitShapePosList instigators, Direction dir,
			double strength) {
		Iterator<UnitShapePosPair> spi;
		UnitShapePosPair spp;
		LinkedList<Position> pl;
		Iterator<Position> pi;
		boolean doable = true;
		double resist;

		// Check instigators actually has a nonempty position:
		pl = instigators.nonEmptyPositions();
		if (pl.isEmpty())
			return 0;

		// first check that the direction is meaningful for all positions
		// in the instigators list
		for (pi = pl.iterator(); pi.hasNext();)
			doable = pi.next().hasNeighbour(dir) && doable;
		if (!doable)
			return 0;

		// build the list of shapes which would get pushed
		UnitShapePosList spl = new UnitShapePosList();
		spl.addAll(instigators);
		abuttingUnitShapes(dir, spl);

		// calculate the total resistance
		resist = Resistance.none;
		for (spi = spl.iterator(); spi.hasNext();)
			resist = resist + spi.next().shape.resistance(dir);
		resist = resist + resistanceFloor_SPL(dir, spl);

		if (resist > strength)
			return 0;
		else {
			Rectangle rect = affectedRectangle(dir, spl);
			// perform a series of ``call-backs'' on the floor to inform it
			// of the move event
			moveFloor_SPL(dir, spl);
			// next move the shapes
			for (spi = spl.iterator(); spi.hasNext();) {
				spp = spi.next();
				spp.pos = spp.pos.neighbour(dir);
			}

			// notify observers
			setChanged();
			notifyObservers(rect);
			return spl.size();
		}
	}

	LinkedList<Position> nonEmptyPositions() {
		LinkedList<Position> pl = unitshape_group.nonEmptyPositions();
		Iterator<Position> pli;
		Position p;
		for (pli = floor_group.nonEmptyPositions().iterator(); pli.hasNext();) {
			p = pli.next();
			if (!pl.contains(p))
				pl.add(p);
		}
		return pl;
	}

	char at(Position pos) {
		return Unit.overlap(unitshape_group.at(pos), floor_group.at(pos));
	}

	/**
	 * adds to the list of shapes, those which would affect a move by the object
	 * at pos in direction dir.
	 */
	private void abuttingUnitShapes(Direction dir, UnitShapePosList uspl) {
		Iterator<Position> pi;
		for (pi = uspl.nonEmptyPositions().iterator(); pi.hasNext();) {
			abuttingUnitShapes_rec(dir, pi.next().neighbour(dir), uspl);
		}
	}

	private void abuttingUnitShapes_rec(Direction dir, Position pos,
			UnitShapePosList uspl) {
		Iterator<Position> pi;
		if (!unitshape_group.empty(pos)) {
			UnitShapePosPair sp = unitshape_group.owner(pos);
			if (!uspl.contains(sp)) {
				uspl.add(sp);
				for (pi = sp.shape.neighbours(sp.pos, dir).iterator(); pi
						.hasNext();)
					abuttingUnitShapes_rec(dir, pi.next(), uspl);
			}
		}
	}

	// works out the rectangular area of the screen affected by a move
	// based on the last provided UnitGraphics object.
	// It might be more efficient to repaint the entire window, but using this
	// probably reduces flicker.
	private Rectangle affectedRectangle(Direction dir, UnitShapePosList spl) {
		// there's always a chance that the level is being used in a context
		// where
		// it's not being redrawn, so we have no RectangleCalculator object.
		// Nevermind, because the return value is only used for drawing.
		if (unit_graphics == null)
			return new Rectangle();
		else {
			Iterator<UnitShapePosPair> spli;
			UnitShapePosPair sp;
			UnitShapeBounds usb;
			if (spl.isEmpty())
				return new Rectangle();
			else {
				spli = spl.iterator();
				sp = spli.next();
				usb = new UnitShapeBounds(sp.shape.unitShapeBounds(), sp.pos);
				usb.updateBoundsOffset(sp.shape.unitShapeBounds(), sp.pos
						.neighbour(dir));
				while (spli.hasNext()) {
					sp = spli.next();
					usb.updateBoundsOffset(sp.shape.unitShapeBounds(), sp.pos);
					usb.updateBoundsOffset(sp.shape.unitShapeBounds(), sp.pos
							.neighbour(dir));
				}
				return unit_graphics.calculateRectangle(usb);
			}
		}
	}

	/**
	 * returns the resistance to the unit at pos moving over onto its
	 * neighbouring floor tile in direction dir.
	 */
	protected double floorResistance(Direction dir, Position pos) {
		return floor_group.resistance(unitshape_group.at(pos), pos, dir);
	}

	/**
	 * This is called when a unit at position pos is moved in direction dir it's
	 * a bit like a callback
	 */
	public void floorMove(Direction dir, Position pos) {
	}

	// returns the cumulative resistance to moving the shapes
	// in spl in direction dir due to the floor
	private double resistanceFloor_SPL(Direction dir, UnitShapePosList spl) {
		double resist = Resistance.none;
		Iterator<Position> pi;
		for (pi = spl.nonEmptyPositions().iterator(); pi.hasNext();)
			resist = resist + floorResistance(dir, pi.next());
		return resist;
	}

	// This iterates over the moving units, calling floorMove for each one
	private void moveFloor_SPL(Direction dir, UnitShapePosList spl) {
		Iterator<Position> pi;
		for (pi = spl.nonEmptyPositions().iterator(); pi.hasNext();)
			floorMove(dir, pi.next());
	}

	public void paintPanel(Graphics2D g2, Insets i, Dimension d) {
		// ensure the current tessellation is correct
		TessellationManager.set_tessellation(get_gameInfo().get_tessellation());
		unit_graphics = new UnitGraphics(g2, i, d, unitShapeBounds());
		paintInto(unit_graphics, TessellationManager.newPosition(),
				UnitShape.unselected);
		unit_graphics.resetGraphics();
	}

	void paintInto(UnitGraphics g, Position rel, int superSelection) {
		// I should probably paint a nice background here
		floor_group.paintInto(g, rel, UnitShape.unselected);
		unitshape_group.paintInto(g, rel, UnitShape.unselected);
	}

	/**
	 * Note that ObservableProxies do not get copied.
	 */
	public Object clone() {
		Level new_instance = (Level) super.clone();
		// shallow
		new_instance.title = title;
		new_instance.copyright = copyright;
		new_instance.floor_group = floor_group;
		// deep
		// new_instance.floor_group = (FloorGroup) floor_group.clone();
		new_instance.unitshape_group = (UnitShapeGroup) unitshape_group.clone();
		new_instance.ob_proxy = new ObservableProxy();
		// if (unit_graphics != null)
		// new_instance.unit_graphics = (UnitGraphics) unit_graphics
		// .clone();
		return new_instance;
	}

	/*
	 * ACCESSORS
	 */

	public String get_title() {
		return title;
	}

	public String get_copyright() {
		return copyright;
	}

	/*
	 * ROUTINES CONCERNING MOUSE CLICKS
	 */

	public Position get_position(Point point) {
		return TessellationManager.newPosition(unit_graphics
				.unTranslatePoint(point));
	}

	/**
	 * Does this position already occur in the routes
	 */

	private boolean visited(Position pos,
			LinkedList<Pair<LinkedList<Position>, LinkedList<Direction>>> routes) {
		boolean seen = false;
		for (Pair<LinkedList<Position>, LinkedList<Direction>> pair : routes) {
			for (Position p : pair.first()) {
				seen = p.equals(pos);
				if (seen)
					break;
			}
			if (seen)
				break;
		}
		return seen;
	}

	public boolean attempt_selectAt(int s, Position pos) {
		// FIXME this is the wrong way to do this.
		UnitShapeBounds usb = unitshape_group.attempt_selectAt(s, pos);
		if (usb != null) {
			if (unit_graphics != null) {
				Rectangle r = unit_graphics.calculateRectangle(usb);
				setChanged();
				notifyObservers(r);
			} else {
				unitshape_group.attempt_selectAt(s, pos);
				setChanged();
				notifyObservers();
			}
			return true;
		} else
			return false;
	}

	/**
	 * Find the shortest route for the selection p to reach position pos or
	 * return null.
	 * 
	 * NOTE: This only works for the current set of games and will not work in
	 * general.
	 */
	public LinkedList<Direction> findRoute(int p, Position pos) {
		// We'll only consider the first non-empty position of the first element
		// of the selected shape.
		// FIXME - this may fail if the selection is empty of has no non-empty
		// positions
		Position startpos = unitshape_group.selected(p).nonEmptyPositions()
				.getFirst();

		// If the start position is clicked, then don't do anything.
		if (startpos.equals(pos))
			return null;

		/*
		 * If a neighbour of the start position is clicked, return a move in
		 * that direction. This behaviour is disabled for triangular
		 * tessellations.
		 */
		if (true) {
			for (DirectionIterator di = new DirectionIterator(startpos); di
					.hasNext();) {
				Direction d = di.next();
				if (startpos.neighbour(d).equals(pos)) {
					LinkedList<Direction> dl = new LinkedList<Direction>();
					dl.add(d);
					return dl;
				}
			}
		}

		/*
		 * While running, the algorithm considers a list of (position list,
		 * direction list)s
		 */
		LinkedList<Pair<LinkedList<Position>, LinkedList<Direction>>> routes = new LinkedList<Pair<LinkedList<Position>, LinkedList<Direction>>>();
		/*
		 * We start with the current position and no directions
		 */
		LinkedList<Position> init_pos = new LinkedList<Position>();
		init_pos.add(startpos);

		routes.add(new Pair<LinkedList<Position>, LinkedList<Direction>>(
				init_pos, new LinkedList<Direction>()));

		/*
		 * Keep iterating until we've consider all routes or we've found a
		 * successful route
		 */
		while (!routes.isEmpty()) {
			Pair<LinkedList<Position>, LinkedList<Direction>> first_route = routes
					.getFirst();
			Position current_pos = first_route.first().getLast();
			if (current_pos.equals(pos))
				break;

			// Extend the route at the front with each position we haven't
			// visited yet
			// First, consider a suitable starting direction
			Direction first_dir = TessellationManager.newDirection();
			first_dir.nextEdgeAt(current_pos);

			// Then, iterate over the directions at that position
			// Direction d = first_dir;
			// boolean first_loop = true;
			// while (first_loop || !d.equals(first_dir)) {
			for (DirectionIterator di = new DirectionIterator(current_pos); di
					.hasNext();) {
				// pl is the new extended list of positions
				Direction d = di.next();
				LinkedList<Position> pl = new LinkedList<Position>(first_route
						.first());
				LinkedList<Direction> dl = new LinkedList<Direction>(
						first_route.second());
				Position newpos = current_pos.neighbour(d);
				// FIXME This emptiness check is not sufficient in general!!!
				if (unitshape_group.empty(newpos) && !visited(newpos, routes)) {
					pl.add(newpos);
					dl.add(d);
					routes
							.add(new Pair<LinkedList<Position>, LinkedList<Direction>>(
									pl, dl));
				}
			}

			// remove the route we've just considered
			routes.removeFirst();
		}

		if (routes.isEmpty())
			return null;
		else
			return routes.getFirst().second();
	}

	/*
	 * WRAPPERS FOR THE OBSERVABLE PROXY
	 */

	public void setChanged() {
		ob_proxy.setChanged();
	}

	public void notifyObservers() {
		ob_proxy.notifyObservers();
	}

	public void notifyObservers(Object o) {
		ob_proxy.notifyObservers(o);
	}

	public void addObserver(Observer o) {
		ob_proxy.addObserver(o);
	}

	public void deleteObserver(Observer o) {
		ob_proxy.deleteObserver(o);
	}
}
