/* The MazezaM Engine 
 * Copyright (C) 2002-2008 Malcolm Tyrrell
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */
package level;

/* UnitGraphics.java -
 *
 * Copyright (C) 2002-2005 Malcolm Tyrrell
 * malcohol@eircom.net
 *
 * This code may be used and distributed under the terms of the GNU General
 * Public Licence.
 */

import globals.SettingsManager;

import java.awt.BasicStroke;
import java.awt.Dimension;
import java.awt.Graphics2D;
import java.awt.Insets;
import java.awt.Paint;
import java.awt.Point;
import java.awt.Rectangle;
import java.awt.Shape;
import java.awt.geom.AffineTransform;
import java.awt.geom.Area;
import java.awt.geom.Ellipse2D;
import java.awt.geom.GeneralPath;
import java.awt.geom.NoninvertibleTransformException;
import java.awt.geom.Point2D;
import java.awt.geom.Rectangle2D;
import java.util.Iterator;
import java.util.LinkedList;

import tessellation.EdgeIterator;
import tessellation.PosDirBoolTuple;
import tessellation.Position;

/**
 * A class which brings together routines for painting UnitShapes.
 */
class UnitGraphics {
	/*
	 * the graphics2D object we're painting into
	 */
	private Graphics2D graphics;

	/*
	 * the affine transform used to scale this level. We will want to be able to
	 * provide this to create a RectangleCalculator. For some reason, we cannot
	 * rely on graphics.getTransform.
	 */
	private AffineTransform aft;

	/**
	 * Constructor takes a Graphics2D object, a y position which is the top
	 * position to start drawing from, a dimension object which describes the
	 * area in pixels, and a position which is the upperbound of the unitShape
	 * we plan to draw.
	 */
	UnitGraphics(Graphics2D g2, Insets i, Dimension dim, UnitShapeBounds usb) {
		double area_width = dim.getWidth() - i.left - i.right;
		double area_height = dim.getHeight() - i.top - i.bottom;
		Point2D topLeft = usb.topLeftCornerPoint();
		Point2D bottomRight = usb.bottomRightCornerPoint();
		double shape_width = bottomRight.getX() - topLeft.getX();
		double shape_height = bottomRight.getY() - topLeft.getY();
		double y_origin = i.top;
		double x_origin = i.left;
		double x_stretch = area_width / shape_width;
		double y_stretch = area_height / shape_height;

		graphics = g2;

		// The main task of the constructor is to create an affinetransform
		// which positions and stretches the painting of the level
		aft = new AffineTransform();

		if (x_stretch * SettingsManager.graphics_Xproportion > y_stretch) {
			// we shouldn't stretch so much in this direction
			// so adjust the stretch to the maximum acceptable value
			x_stretch = y_stretch / SettingsManager.graphics_Xproportion;
			x_origin = x_origin + (area_width - (shape_width * x_stretch))
					/ 2.0;
		} else if (y_stretch * SettingsManager.graphics_Yproportion > x_stretch) {
			y_stretch = x_stretch / SettingsManager.graphics_Yproportion;
			y_origin = y_origin + (area_height - (shape_height * y_stretch))
					/ 2.0;
		}

		aft.translate(x_origin, y_origin);

		aft.scale(x_stretch, y_stretch);

		// Because the origin of the level is not necessarily the topleftmost
		// unit in the shape, we have to translate our drawing accordingly
		aft.translate(-topLeft.getX(), -topLeft.getY());

		g2.transform(aft);

		// as we are scaling the graphics, the default thickness of a line is
		// likely to be too big. We adjust it back to about 1 pixels.
		g2.setStroke(new BasicStroke((float) (2.0 / (x_stretch + y_stretch))));
	}

	/**
	 * Reset the given graphics2d object.
	 */
	// FIXME This can't be the right way to achieve this effect
	void resetGraphics() {
		try {
			graphics.transform(aft.createInverse());
			graphics.setStroke(new BasicStroke(1.0f));
		} catch (NoninvertibleTransformException e) {
			// The transforms should always be invertible
			throw new Error ("The transform is uninvertible");
		}
	}
	
	void setPaint(Paint p) {
		graphics.setPaint(p);
	}

	void draw(Shape s) {
		graphics.draw(s);
	}

	void fill(Shape s) {
		graphics.fill(s);
	}

	/**
	 * Return the point resulting from the inverse transform.
	 */
	Point2D unTranslatePoint(Point2D point) {
		try {
			return aft.inverseTransform(point, null);
		} catch (NoninvertibleTransformException e) {
			// The transforms should always be invertible
			throw new Error ("The transform is uninvertible");
		}
	}

	/**
	 * Returns an circle which fits inside the provided position
	 */
	Ellipse2D unitCircle(Position pos) {
		Point2D topLeft = pos.innerCircleTopLeftCornerPoint();
		Point2D bottomRight = pos.innerCircleBottomRightCornerPoint();
		double x = topLeft.getX();
		double y = topLeft.getY();
		return new Ellipse2D.Double(x, y, bottomRight.getX() - x, bottomRight
				.getY()
				- y);
	}

	/**
	 * Return an area which is a circle with a hole in it.
	 */
	Area unitCircleHole(Position pos) {
		Point2D topLeft = pos.innerCircleTopLeftCornerPoint();
		Point2D bottomRight = pos.innerCircleBottomRightCornerPoint();
		double x = topLeft.getX();
		double y = topLeft.getY();
		double w = bottomRight.getX() - x;
		double h = bottomRight.getY() - y;
		double c_h = w * 0.1;
		double c_v = h * 0.1;
		Area a = new Area(new Ellipse2D.Double(x, y, w, h));
		a.subtract(new Area(new Ellipse2D.Double(x + c_h, y + c_v, w
				- (2.0 * c_h), h - (2.0 * c_v))));
		return a;
	}

	/**
	 * Returns an square which fits inside the provided position
	 */
	Rectangle2D unitSquare(Position pos) {
		Point2D topLeft = pos.innerSquareTopLeftCornerPoint();
		Point2D bottomRight = pos.innerSquareBottomRightCornerPoint();
		double x = topLeft.getX();
		double y = topLeft.getY();
		return new Rectangle2D.Double(x, y, bottomRight.getX() - x, bottomRight
				.getY()
				- y);
	}

	/**
	 * Returns a path around the given unit
	 */
	GeneralPath unitOutline(Position pos) {
		LinkedList<PosDirBoolTuple> pdl = new LinkedList<PosDirBoolTuple>();
		EdgeIterator ei = new EdgeIterator(pos);
		ei.hasNext();
		while (ei.hasNext()) {
			pdl.add(new PosDirBoolTuple(pos, ei.next(), true));
		}
		return complexOutline(pdl);
	}

	/**
	 * Returns a path around the given unit
	 */
	GeneralPath unitInnerOutline(Position pos) {
		LinkedList<PosDirBoolTuple> pdl = new LinkedList<PosDirBoolTuple>();
		EdgeIterator ei = new EdgeIterator(pos);
		ei.hasNext();
		while (ei.hasNext()) {
			pdl.add(new PosDirBoolTuple(pos, ei.next(), true));
		}
		return complexInnerOutline(pdl);
	}

	/**
	 * The first edge path defines the outer outline, the others define the
	 * holes if there are any.
	 */
	Area edgePathWithHoles(OutlinePair op) {
		Iterator<LinkedList<PosDirBoolTuple>> pdlli;

		// first, create the outer outline
		Area a = new Area(complexOutline(op.outer_outline));

		// next, remove the holes if there are any
		for (pdlli = op.the_holes.iterator(); pdlli.hasNext();) {
			a.subtract(new Area(complexOutline(pdlli.next())));
		}

		return a;
	}

	/**
	 * The first edge path defines the outer outline, the others define the
	 * holes if there are any.
	 */
	Area innerEdgePathWithHoles(OutlinePair op) {
		Iterator<LinkedList<PosDirBoolTuple>> pdlli;

		// first, create the outer outline
		Area a = new Area(complexInnerOutline(op.outer_outline));

		// next, remove the holes if there are any
		for (pdlli = op.the_holes.iterator(); pdlli.hasNext();) {
			a.subtract(new Area(complexInnerOutline(pdlli.next())));
		}

		return a;
	}

	GeneralPath complexOutline(LinkedList<PosDirBoolTuple> pdl) {
		GeneralPath gp = new GeneralPath();
		Iterator<PosDirBoolTuple> pdli = pdl.iterator();
		PosDirBoolTuple pdp;
		Point2D startPoint;
		if (!pdli.hasNext()) {
			throw new Error("Empty edge vector list supplied");
		}
		pdp = pdli.next();
		startPoint = pdp.pos.startPointOfEdge(pdp.dir);
		gp.moveTo((float) startPoint.getX(), (float) startPoint.getY());
		while (pdli.hasNext()) {
			pdp = pdli.next();
			if (pdp.bool) {
				startPoint = pdp.pos.startPointOfEdge(pdp.dir);
				gp.lineTo((float) startPoint.getX(), (float) startPoint.getY());
			}
		}
		gp.closePath();
		return gp;
	}

	GeneralPath complexInnerOutline(LinkedList<PosDirBoolTuple> pdl) {
		GeneralPath gp = new GeneralPath();
		Iterator<PosDirBoolTuple> pdli = pdl.iterator();
		PosDirBoolTuple pdp;
		Point2D startPoint, endPoint;
		if (!pdli.hasNext()) {
			throw new Error("Empty edge vector list supplied");
		}
		pdp = pdli.next();
		startPoint = pdp.pos.startPointOfInnerEdge(pdp.dir);
		gp.moveTo((float) startPoint.getX(), (float) startPoint.getY());
		endPoint = pdp.pos.endPointOfInnerEdge(pdp.dir);
		gp.lineTo((float) endPoint.getX(), (float) endPoint.getY());
		while (pdli.hasNext()) {
			pdp = pdli.next();
			startPoint = pdp.pos.startPointOfInnerEdge(pdp.dir);
			gp.lineTo((float) startPoint.getX(), (float) startPoint.getY());
			if (pdp.bool) {
				endPoint = pdp.pos.endPointOfInnerEdge(pdp.dir);
				gp.lineTo((float) endPoint.getX(), (float) endPoint.getY());
			}
		}
		gp.closePath();
		return gp;
	}

	Area edgeOutlineOutline(OutlinePair op) {
		Area a = edgePathWithHoles(op);
		a.subtract(innerEdgePathWithHoles(op));
		return a;
	}

	Area unitOutlineOutline(Position pos) {
		Area a = new Area(unitOutline(pos));
		a.subtract(new Area(unitInnerOutline(pos)));
		return a;
	}

	/**
	 * Returns the Rectangle defined by the UnitShapeBounds object
	 */
	Rectangle calculateRectangle(UnitShapeBounds usb) {
		if (usb == null) {
			return null;
		}
		Point2D p0 = usb.topLeftCornerPoint();
		Point2D p1 = usb.bottomRightCornerPoint();
		// transform the points according to the affine transform
		p0 = aft.transform(p0, null);
		p1 = aft.transform(p1, null);
		// FIXME is this the only way of providing a rectangle with double
		// coordinates.
		Point p = new Point();
		p.setLocation(p0.getX() - 1.0, p0.getY() - 1.0);
		Dimension d = new Dimension();
		d.setSize(p1.getX() - p0.getX() + 2.0, p1.getY() - p0.getY() + 2.0);
		return new Rectangle(p, d);
	}
}