/* The MazezaM Engine 
 * Copyright (C) 2002-2008 Malcolm Tyrrell
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */
package tessellation;

import java.awt.geom.Point2D;

/**
 * A position in the Square Tessellation.
 */
class SquarePosition extends Position {
	/**
	 * Default constructor returns the origin
	 */
	SquarePosition() {
		x = 0;
		y = 0;
	}

	/**
	 * constructor which takes array coordinates and constructs the
	 * corresponding position
	 */
	SquarePosition(int array_x, int array_y) {
		x = array_x;
		y = array_y;
	}

	/**
	 * Constructor clones the provided position
	 */
	SquarePosition(Position pos) {
		x = pos.x;
		y = pos.y;
	}

	SquarePosition(Point2D p) {
		x = (int) p.getX();
		y = (int) p.getY();
	}

	/**
	 * Constructs the position which neighbours this position in the provided
	 * direction
	 */
	public Position neighbour(Direction dir) {
		if (dir.is(SquareDirection.north))
			return new SquarePosition(x, y - 1);
		else if (dir.is(SquareDirection.east))
			return new SquarePosition(x + 1, y);
		else if (dir.is(SquareDirection.south))
			return new SquarePosition(x, y + 1);
		else
			return new SquarePosition(x - 1, y);
	}

	/**
	 * Returns a position which corresponds to adding the provided positions
	 * together.
	 */
	public Position plus(Position rel) {
		return TessellationManager.newPosition(x + rel.x, y + rel.y);
	}

	/**
	 * Calculates the a position which is the relative difference between two
	 * positions
	 */
	public Position minus(Position rel) {
		return TessellationManager.newPosition(x - rel.x, y - rel.y);
	}

	/**
	 * Returns the position which extends furthest up the screen
	 */
	public Position furthestUp(Position other) {
		return (y <= other.y) ? this : other;
	}

	/**
	 * Returns the position which extends furthest right on the screen
	 */
	public Position furthestRight(Position other) {
		return (x >= other.x) ? this : other;
	}

	/**
	 * Returns the position which extends furthest down on the screen
	 */
	public Position furthestDown(Position other) {
		return (y >= other.y) ? this : other;
	}

	/**
	 * Returns the position which extends furthest left on the screen
	 */
	public Position furthestLeft(Position other) {
		return (x <= other.x) ? this : other;
	}

	/**
	 * returns the coordinates for the start of the given edge of the given
	 * position
	 */
	public Point2D startPointOfEdge(Direction dir) {
		if (dir.is(SquareDirection.north))
			return new Point2D.Double(x, y);
		else if (dir.is(SquareDirection.east))
			return new Point2D.Double(x + 1, y);
		else if (dir.is(SquareDirection.south))
			return new Point2D.Double(x + 1, y + 1);
		else if (dir.is(SquareDirection.west))
			return new Point2D.Double(x, y + 1);
		else
			throw new Error("Invalid direction provided");
	}

	/**
	 * returns the coordinates of the upperlefthand corner which bounds the
	 * given position
	 */
	public Point2D topLeftCornerPoint() {
		return new Point2D.Double(x, y);
	}

	/**
	 * returns the coordinates of the lowerrighthand corner which bounds the
	 * given position
	 */
	public Point2D bottomRightCornerPoint() {
		return new Point2D.Double(x + 1, y + 1);
	}

	/**
	 * returns the coordinates of the upperlefthand corner of the unit's inner
	 * circle
	 */
	public Point2D innerCircleTopLeftCornerPoint() {
		return new Point2D.Double(x + 0.05, y + 0.05);
	}

	/**
	 * returns the coordinates of the lowerrighthand corner of the unit's inner
	 * circle
	 */
	public Point2D innerCircleBottomRightCornerPoint() {
		return new Point2D.Double(x + 0.95, y + 0.95);
	}

	/**
	 * returns the coordinates of the upperlefthand corner of the unit's inner
	 * square
	 */
	public Point2D innerSquareTopLeftCornerPoint() {
		return new Point2D.Double(x + 0.1, y + 0.1);
	}

	/**
	 * returns the coordinates of the lowerrighthand corner of the unit's inner
	 * square
	 */
	public Point2D innerSquareBottomRightCornerPoint() {
		return new Point2D.Double(x + 0.9, y + 0.9);
	}
}
